from os import path

class Exemptions:
	# Contents of the similarity exemptions file (a list of TwoPartExemption objects)
	similarity_exemptions = []
	# Contents of the lower case exemptions file (a list of strings)
	lower_case_exemptions = []
	# Contents of the reversal exemptions file (a list of TwoPartExemption objects)
	reversal_exemptions = []
	# Contents of the "the" exemptions file (a list of strings)
	the_exemptions=set([])

	def __init__(self,config):
		self.reversal_exemptions=get_items_from_data_text_file(config.paths.reversal_exemptions,parse_reversal_exemption)
		self.the_exemptions=set(get_items_from_data_text_file(config.paths.the_exemptions,parse_simple_exemption))
		self.similarity_exemptions=get_items_from_data_text_file(config.paths.similarity_exemptions,parse_similarity_exemption)
		self.lower_case_exemptions=get_items_from_data_text_file(config.paths.lower_case_exemptions,parse_simple_exemption)

	def is_exempt_from_reversal_check(self,artist1, artist2):
		for reversalExemption in self.reversal_exemptions:
			if reversalExemption.matches(artist1, artist2):
				return True
		return False

	def is_exempt_from_lower_case_check(self,artist1):
		for lowerCaseExemption in self.lower_case_exemptions:
			if lowerCaseExemption in artist1:
				return True
		return False

	def is_exempt_from_similarity_check(self,title1, title2):
		for similarityExemption in self.similarity_exemptions:
			if similarityExemption.matches(title1, title2):
				return True
		return False

	def is_exempt_from_the_check(self,artist1):
		return artist1 in self.the_exemptions

class TwoPartExemption:
	part1 = None
	part2 = None

	def __init__(self, part1, part2):
		self.part1 = part1
		self.part2 = part2

	def matches(self, part1, part2):
		return (self.part1 == part1 and self.part2 == part2) or (self.part1 == part2 and self.part2 == part1)

def parse_two_part_exemption(line, creatorFunc, exemptionType):
	line = line.strip()
	if len(line) > 0:
		bits = line.split("\t")
		if len(bits) == 2:
			return creatorFunc(bits[0], bits[1])
		else:
			raise Exception(f"Could not parse {exemptionType} exemption: {line}")

def create_two_part_exemption(part1,part2):
	return TwoPartExemption(part1,part2)

def parse_reversal_exemption(line):
	return parse_two_part_exemption(line, create_two_part_exemption, "reversal")

def parse_similarity_exemption(line):
	return parse_two_part_exemption(line, create_two_part_exemption, "similarity")

def parse_simple_exemption(line):
	return line.strip()

def read_lines_from_data_text_file(full_path):
	if path.isfile(full_path):
		with open(full_path, mode="r", encoding="utf-8") as f:
			return f.readlines()
	return []

def get_items_from_data_text_file(path,parserFunction):
	lines=read_lines_from_data_text_file(path)
	items=list(map(lambda line: parserFunction(line),lines))
	items=list(filter(lambda item: not item is None, items))
	return items


from enum import Enum, auto
from colorama import Fore, Style

# List of commands as enums
class Severity(Enum):
	ERROR = auto()
	INFO = auto()

class Message:
	message=None
	severity=Severity.ERROR
	def __init__(self,message,severity):
		self.message=message
		self.severity=severity

	def print(self):
		if self.severity==Severity.INFO:
			print(f'{Fore.YELLOW}')
		else:
			print(f'{Fore.RED}')
		print(f'{Style.BRIGHT}{self.message}{Style.RESET_ALL}')

class Error(Message):
	def __init__(message):
		Message(message,Severity.ERROR)

class Info(Message):
	def __init__(message):
		Message(message,Severity.INFO)


from collections import defaultdict
from colorama import Fore, Style
from textdistance import levenshtein
from os import path, walk
from time import sleep
from karaokemanager.display_functions import pad_or_ellipsize
from karaokemanager.karaoke_file import KaraokeFile
from karaokemanager.music_file import MusicFile
from karaokemanager.error import Error
from enum import Enum, auto

# How deep should the library analysis be?
class LibraryAnalysisType(Enum):
	NONE = auto()
	QUICK = auto()
	FULL = auto()

class Library:
	# Current background music playlist (strings from the BackgroundMusicPlaylist file)
	background_music_playlist = set([])
	# List of karaoke files (KaraokeFile objects)
	karaoke_files = []
	# List of music files (MusicFile objects)
	music_files = []
	# Dictionary of karaoke files. Key is artist, value is another dictionary of karaoke files
	# where the key is the song title, and the value is a list of tracks from one or more vendors.
	karaoke_dictionary = defaultdict()
	# Dictionary of music files. Key is artist, value is another dictionary of music files
	# where the key is the song title, and the value is a list of tracks with that title.
	music_dictionary = defaultdict()
	# Files that were ignored
	ignored_files=[]
	# Full manifest of BGM paths
	bgm_manifest=[]
	# All our lovely exemption info
	exemptions=None
	# BGM playlist entries that did not match a music file.
	missing_bgm_playlist_entries=[]
	# Filenames that matched an extension of a pattern, but failed the regex.
	unparseable_filenames=[]
	# Duplicates found during a scan
	karaoke_duplicates=[]
	music_duplicates=[]
	# Analysis errors found during a quick or full analysis of the fileset
	karaoke_analysis_results=[]
	# Analysis errors found during a quick or full analysis of the fileset
	music_analysis_results=[]

	def __init__(self,exemptions):
		self.exemptions = exemptions

	# Reads the background music playlist into memory
	def get_background_music_playlist(self, config):
		playlist = set([])
		backgroundMusicFilePath=config.paths.bgm_playlist
		if path.isfile(backgroundMusicFilePath):
			with open(backgroundMusicFilePath, mode="r", encoding="utf-8") as f:
				lines = f.readlines()
				for line in lines:
					line = line.strip()
					if(len(line) > 0):
						playlist.add(line.lower())
		return playlist

	# Builds the dictionaries of karaoke and music tracks
	def build_dictionaries(self):
		self.karaoke_dictionary = defaultdict()
		for songFile in self.karaoke_files:
			self.karaoke_dictionary.setdefault(songFile.artist, defaultdict()).setdefault(
				songFile.title, []).append(songFile)
		self.music_dictionary = defaultdict()
		for songFile in self.music_files:
			self.music_dictionary.setdefault(songFile.artist, defaultdict()).setdefault(
				songFile.title, []).append(songFile)

	# Builds the karaoke and music lists by analysing folder contents.
	def build_song_lists(self, config, analysis_type, console_size, feedback):
		bgm_playlist=self.get_background_music_playlist(config)

		def create_karaoke_file(path, group_map):
			vendor=group_map["vendor"]
			if vendor is None:
				vendor="UNKNOWN"
			return KaraokeFile(path,group_map["artist"],group_map["title"],vendor)

		def create_music_file(path, group_map):
			return MusicFile(path,group_map["artist"],group_map["title"])

		def parse_filename(file_path, filename, patterns, file_builder):
			name_without_extension, extension = path.splitext(filename)
			extension=extension.strip('.')
			valid_patterns=list(filter(lambda pattern: pattern.extension_matches(extension), patterns))
			if any(valid_patterns):
				for pattern in valid_patterns:
					group_map = pattern.parse_filename(name_without_extension)
					if any(group_map):
						file=file_builder(file_path, group_map)
						if not file is None:
							return file
				self.unparseable_filenames.append(filename)
			else:
				self.ignored_files.append(filename)
			return None

		# Tries to parse a karaoke file, adding it to a collection if successful.
		def scan_karaoke_file(root, file):
			karaokeFile = parse_filename(path.join(root,file), file, config.karaoke_patterns, create_karaoke_file)
			if not karaokeFile is None:
				self.karaoke_files.append(karaokeFile)

		# Tries to parse a music file, adding it to a collection if successful.
		def scan_music_file(root, file):
			full_path = path.join(root,file)
			music_file = parse_filename(full_path, file, config.music_patterns, create_music_file)
			if not music_file is None:
				name_without_extension, _ = path.splitext(file)
				name_without_extension=name_without_extension.lower()
				if name_without_extension in bgm_playlist:
					self.bgm_manifest.append(full_path)
					bgm_playlist.remove(name_without_extension)
				self.music_files.append(music_file)

		# Scans the files in one or more folders.
		def scan_files(file_paths, scan_file_function):
			for filePath in file_paths:
				for root, _, files in walk(filePath):
					print(pad_or_ellipsize(f"Scanning {root}", console_size.columns), end="\r")
					for file in files:
						scan_file_function(root,file)

		# Writes a list of strings to a text file.
		def write_text_file(item_list,path,errors):
			try:
				with open(path, mode="w", encoding="utf-8") as f:
					for item in item_list:
						f.writelines(f"{item}\n")
			except PermissionError:
				errors.append(Error(f"Failed to write to '{path}'."))

		self.bgm_manifest=[]
		self.ignored_files=[]
		self.karaoke_files=[]
		self.music_files=[]
		self.unparseable_filenames=[]
		scan_files(config.paths.karaoke,scan_karaoke_file)
		scan_files(config.paths.music,scan_music_file)
		# Whatever's left in the background music playlist will be missing files.
		self.missing_bgm_playlist_entries=bgm_playlist

		self.build_dictionaries()

		self.music_duplicates=[]
		self.karaoke_duplicates=[]
		self.karaoke_analysis_results=[]
		self.music_analysis_results=[]
		if analysis_type != LibraryAnalysisType.NONE:
			print(pad_or_ellipsize(f"Analyzing karaoke files...", console_size.columns))
			self.karaoke_duplicates, self.karaoke_analysis_results=self.analyze_file_set(self.karaoke_files,self.karaoke_dictionary, analysis_type, console_size)
			print(pad_or_ellipsize(f"Analyzing music files...", console_size.columns))
			self.music_duplicates, self.music_analysis_results=self.analyze_file_set(self.music_files,self.music_dictionary, analysis_type, console_size)

		write_text_file(self.unparseable_filenames,config.paths.filename_errors,feedback)
		write_text_file(self.ignored_files,config.paths.ignored_files,feedback)
		write_text_file(bgm_playlist,config.paths.missing_playlist_entries,feedback)
		write_text_file(self.bgm_manifest,config.paths.bgm_manifest,feedback)
		write_text_file(self.karaoke_duplicates,config.paths.karaoke_duplicates,feedback)
		write_text_file(self.music_duplicates,config.paths.music_duplicates,feedback)
		write_text_file(self.karaoke_analysis_results,config.paths.karaoke_analysis_results,feedback)
		write_text_file(self.music_analysis_results,config.paths.music_analysis_results,feedback)

		# Helper function for dictionary sorting.
		def get_music_file_key(file):
			return file.artist
		self.music_files.sort(key=get_music_file_key)
		self.karaoke_files.sort(key=get_music_file_key)

	# Scans a list of files for potential duplicates, bad filenames, etc.
	def analyze_file_set(self,files,dictionary,analysis_type, console_size):
		analysis_results=[]
		duplicates=[]
		# Checks two strings for similarity.
		def similarity(s1, s2):
			longer_length = max(len(s1),len(s2))
			if longer_length == 0:
				return 1.0
			return (longer_length - levenshtein(s1, s2)) / longer_length

		artists = set([])
		artist_list = []
		artist_lower_list = []

		last_percent = -1
		counter = 0
		song_progress_count = len(dictionary)
		for artist, song_dict in dictionary.items():
			counter += 1
			percent = round((counter/song_progress_count)*100.0)
			if percent > last_percent:
				print(pad_or_ellipsize(f"Looking for duplicates: {percent}% done", console_size.columns), end="\r")
				last_percent = percent
			for song_collection in song_dict.values():
				if len(song_collection)>1:
					duplicates.extend(map(lambda song: f"{song.artist} - {song.title}", song_collection[1:]))
		for song in files:
			if not song.artist in artists:
				artists.add(song.artist)
				artist_list.append(song.artist)
				artist_lower_list.append(song.lower_artist)
		for artist in artists:
			first_letter = artist[0]
			if first_letter.isalpha() and first_letter.islower():
				if not self.exemptions.is_exempt_from_lower_case_check(artist):
					error = f"Artist \"{artist}\" is not capitalised."
					analysis_results.append(error)
			if artist.startswith("The "):
				if artist[4:] in artists and not self.exemptions.is_exempt_from_the_check(artist):
					error = f"Artist \"{artist}\" has a non-The variant."
					analysis_results.append(error)
		artist_count = len(artist_list)
		song_count = len(files)
		song_progress_count = round((song_count*song_count)/2)
		artist_progress_count = round((artist_count*artist_count)/2)
		counter = 0
		last_percent = -1
		for i in range(0, artist_count):
			artist = artist_list[i]
			artist_lower = artist_lower_list[i]
			ampersand_first_index = artist.find(" & ")
			ampersand_last_index = artist.rfind(" & ")
			if ampersand_first_index != -1 and ampersand_last_index == ampersand_first_index:
				bit1 = artist[0:ampersand_first_index]
				bit2 = artist[ampersand_first_index+3:]
				if bit2 != bit1:
					if not self.exemptions.is_exempt_from_reversal_check(bit1, bit2):
						reverse_check = bit2+" & "+bit1
						if reverse_check in artists:
							error = f"Artist \"{artist}\" also appears as \"{reverse_check}\"."
							analysis_results.append(error)
			for j in range(i+1, artist_count):
				counter += 1
				percent = round((counter/artist_progress_count)*100.0)
				if percent > last_percent:
					print(pad_or_ellipsize(f"Analyzing artists: {percent}% done", console_size.columns), end="\r")
					last_percent = percent
				compare_artist = artist_list[j]
				compare_artist_lower = artist_lower_list[j]
				if artist_lower == compare_artist_lower and artist != compare_artist:
					error = f"Artist \"{artist}\" has a case variation: \"{compare_artist}\"."
					analysis_results.append(error)
		song_progress_count = round((song_count*song_count)/2)
		counter = 0
		last_percent = -1
		for i in range(0, song_count):
			song_file = files[i]
			song_title = song_file.title
			song_title_lower = song_file.lower_title
			first_letter = song_title[0]
			if first_letter.isalpha() and first_letter.islower():
				if not self.exemptions.is_exempt_from_lower_case_check(song_title):
					error = f"Title \"{song_title}\" is not capitalised."
					analysis_results.append(error)
			for j in range(i+1, song_count):
				counter += 1
				percent = round((counter/song_progress_count)*100.0)
				if percent > last_percent:
					print(pad_or_ellipsize(f"Analyzing song titles (simple analysis): {percent}% done", console_size.columns), end="\r")
					last_percent = percent
				compare_title = files[j].title
				compare_title_lower = files[j].lower_title
				if song_title != compare_title:
					if song_title_lower == compare_title_lower:
						error = f"Title \"{song_title}\" has a case variation: \"{compare_title}\"."
						analysis_results.append(error)
		last_percent = -1
		counter = 0
		song_progress_count = len(dictionary)
		if LibraryAnalysisType.FULL == analysis_type:
			for artist, song_dict in dictionary.items():
				counter += 1
				percent = round((counter/song_progress_count)*100.0)
				if percent > last_percent:
					print(pad_or_ellipsize(f"Analyzing song titles (complex analysis): {percent}% done", console_size.columns), end="\r")
					last_percent = percent
				keys = list(song_dict.keys())
				for i in range(0, len(keys)):
					for j in range(i+1, len(keys)):
						if not self.exemptions.is_exempt_from_similarity_check(keys[i], keys[j]):
							similarity_calculation = similarity(keys[i], keys[j])
							if similarity_calculation < 1.0 and similarity_calculation > 0.9:
								error = f"Title \"{keys[i]}\" looks very similar to \"{keys[j]}\"."
								analysis_results.append(error)
		return duplicates, analysis_results


import yaml
from os import path, makedirs, remove
import importlib
from karaokemanager.file_pattern import FilePattern

class Config:
	karaoke_patterns=[]
	music_patterns=[]
	paths = None
	driver = None

	def __init__(self, config_path):
		if path.isfile(config_path):
			config = yaml.safe_load(open(config_path))
			if not config is None:
				self.paths=Paths(config)

				self.karaoke_patterns=get_patterns(config,"karaoke")
				if not any(self.karaoke_patterns):
					raise Exception("No karaoke file patterns defined.")
				self.music_patterns=get_patterns(config,"music")
				if any(self.paths.music) and not any(self.music_patterns):
					raise Exception("No music file patterns defined.")

				self.driver=create_driver(config)
				return
			raise Exception("Failed to parse YAML configuration.")
		raise Exception(f"{config_path} not found.")

def create_driver(config):
	driverConfig=config.get("driver")
	if driverConfig is None:
		raise Exception("No 'driver' section in configuration file.")
	else:
		driverClassString=get_config_string(driverConfig,"class",True)
		driverSpecificConfig=driverConfig.get(driverClassString)
		if driverSpecificConfig is None:
			raise Exception("No driver-specific config section found in configuration file.")
		else:
			components = driverClassString.split('.')
			if len(components)<3:
				raise Exception("Driver class string should be in the format 'packagename.modulename.classname'")
			else:
				upperBound=len(components)-1
				package=components[0]
				module=".".join(components[1:upperBound])
				className=components[upperBound]
				module = importlib.import_module(module, package)
				driverClass = getattr(module, className)
				return driverClass(driverSpecificConfig)

def get_config_string(config, name, required):
	if name in config:
		value=config.get(name)
		if isinstance(value, str):
			value=value.strip()
			if len(value)>0:
				return value
	if required:
		raise Exception(f"No string value found for '{name}' in YAML config file.")
	return None

def get_patterns(config,section):
	patterns=[]
	if section in config:
		section_config=config[section]
		if "patterns" in section_config:
			unparsed_patterns=section_config["patterns"]
			patterns=list(map(lambda unparsed_pattern: parse_file_pattern(unparsed_pattern), unparsed_patterns))
	return patterns

def get_paths(config,section,required):
	paths=[]
	if section in config:
		section_config=config[section]
		if "paths" in section_config:
			paths=section_config["paths"]
	if required and not any(paths):
		raise Exception(f"No paths defined in '{section}' section")
	return paths

def is_string_list(l):
	if isinstance(l,list):
		return all(map(lambda item: isinstance(item, str), l))
	return False
	
def parse_file_pattern(pattern_config):
	regular_expression=get_config_string(pattern_config,"regExp",True)
	group_names=pattern_config.get("groupNames")
	extensions=pattern_config.get("extensions")
	if group_names is None or not is_string_list(group_names):
		raise Exception("No groupNames defined for pattern.")
	if extensions is None or not is_string_list(extensions):
		raise Exception("No extensions defined for pattern.")
	return FilePattern(extensions,regular_expression,group_names)
	
class Paths:
	# Filename of the background music list file (just artist & titles)
	BACKGROUND_MUSIC_PLAYLIST_FILENAME = "BackgroundMusicPlaylist.txt"
	# Filename of the background music manifest (full paths, calculated from playlist)	
	BACKGROUND_MUSIC_MANIFEST_FILENAME = "KaraokeManager.backgroundMusic.txt"
	# Filename of the requests file that we can write to.
	MUSIC_REQUESTS_FILENAME="KaraokeManager.musicRequests.txt"
	# Name of file that we will write to when we find malformed filenames.
	FILENAME_ERRORS_FILENAME = "BadFilenames.txt"
	# Name of file that we will write to when we find BackgroundMusicPlaylist entries that do not correspond
	# to a known music file.
	MISSING_PLAYLIST_ENTRIES_FILENAME = "MissingPlaylistEntries.txt"
	# Name of file that we will write to when we find possible duplicates of music files.
	MUSIC_DUPLICATES_FILENAME = "MusicDuplicates.txt"
	# Name of file that we will write to when we find possible duplicates of karaoke files.
	KARAOKE_DUPLICATES_FILENAME = "KaraokeDuplicates.txt"
	# Name of file that we will write to when we find karaoke files with artist and/or title problems.
	KARAOKE_ANALYSIS_RESULTS_FILENAME = "KaraokeAnalysisResults.txt"
	# Name of file that we will write to when we find music files with artist and/or title problems.
	MUSIC_ANALYSIS_RESULTS_FILENAME = "MusicAnalysisResults.txt"
	# Name of file that we will write to with a list of files that we ignored
	IGNORED_FILES_FILENAME = "IgnoredFiles.txt"
	# File that we write a random song suggestion to periodically.
	SONG_SUGGESTION_FILENAME="KaraokeManager.songSuggestion.txt"
	# Filename to persist the state to.
	# Gets written after every command, in case the app crashes.
	STATE_FILENAME = "KaraokeManager.state.txt"
	# Filename to persist the list of singers to.
	# Gets written after every command, in the case the app crashes.
	# Gets read by the Winamp SingersQueue plugin to show the current queue onscreen.
	SINGERS_QUEUE_FILENAME = "KaraokeManager.singers.txt"
	# List of reversal exemptions. If we find a music file by "Chas & Dave", and one by "Dave & Chas",
	# this will be flagged as a problem unless there is a line in the ReversalExemptions file
	# declaring this reversal to be exempt from challenge. Each line in this file should be two
	# string separated by a tab, for example "Chas\tDave".
	REVERSAL_EXEMPTIONS_FILENAME = "ReversalExemptions.txt"
	# List of "the" exemptions. Usually, if we find a music file by an artist/group that also exists
	# prefixed by "The" (e.g. one file by "Rolling Stones" and one by "The Rolling Stones") then this
	# will be flagged as a problem, unless an exemption is listed in this file. Either of the two
	# values can be used.
	THE_EXEMPTIONS_FILENAME = "TheExemptions.txt"
	# Normally, we want artist names and song titles to be capitalised. However, some annoyances
	# exist (e.g. "k.d. lang"). List them in this file to make them exempt from challenge.
	LOWER_CASE_EXEMPTIONS_FILENAME = "LowerCaseExemptions.txt"
	# If we find two artist names or song titles that look INCREDIBLY similar, we flag them as
	# potential duplicates with an unintentional typo. However, sometimes this yields false positives.
	# List them in this file to get rid of them, with both values separated by a tab.
	SIMILARITY_EXEMPTIONS_FILENAME = "SimilarityExemptions.txt"

	requests=None
	bgm_manifest=None
	bgm_playlist=None
	random_suggestion=None
	karaoke_duplicates=None
	music_duplicates=None
	filename_errors=None
	missing_playlist_entries=None
	karaoke_analysis_results=None
	music_analysis_results=None
	ignored_files=None
	reversal_exemptions=None
	the_exemptions=None
	lower_case_exemptions=None
	similarity_exemptions=None

	def __init__(self, config):
		data_path=get_config_string(config,'dataPath',True)
		temp_data_path=get_config_string(config,'tempDataPath',True)

		if not path.isdir(temp_data_path):
			makedirs(temp_data_path)
		if not path.isdir(data_path):
			makedirs(data_path)

		bgm_playlist_filename=self.BACKGROUND_MUSIC_PLAYLIST_FILENAME
		musicConfig=config.get("music")
		if not musicConfig is None:
			temp = get_config_string(musicConfig, "backgroundMusicPlaylistFilename", False)
			if not temp is None:
				bgm_playlist_filename=temp

		self.requests = path.join(temp_data_path, self.MUSIC_REQUESTS_FILENAME)
		self.bgm_playlist = path.join(data_path,bgm_playlist_filename)
		self.bgm_manifest = path.join(temp_data_path, self.BACKGROUND_MUSIC_MANIFEST_FILENAME)
		self.music_duplicates=path.join(data_path,self.MUSIC_DUPLICATES_FILENAME)
		self.karaoke_duplicates=path.join(data_path,self.KARAOKE_DUPLICATES_FILENAME)
		self.music_analysis_results=path.join(data_path,self.MUSIC_ANALYSIS_RESULTS_FILENAME)
		self.karaoke_analysis_results=path.join(data_path,self.KARAOKE_ANALYSIS_RESULTS_FILENAME)
		self.random_suggestion = path.join(temp_data_path, self.SONG_SUGGESTION_FILENAME)
		self.filename_errors=path.join(data_path,self.FILENAME_ERRORS_FILENAME)
		self.missing_playlist_entries=path.join(data_path,self.MISSING_PLAYLIST_ENTRIES_FILENAME)
		self.ignored_files=path.join(data_path,self.IGNORED_FILES_FILENAME)
		self.state=path.join(temp_data_path,self.STATE_FILENAME)
		self.singers_queue=path.join(temp_data_path,self.SINGERS_QUEUE_FILENAME)
		self.reversal_exemptions=path.join(data_path,self.REVERSAL_EXEMPTIONS_FILENAME)
		self.the_exemptions=path.join(data_path,self.THE_EXEMPTIONS_FILENAME)
		self.lower_case_exemptions=path.join(data_path,self.LOWER_CASE_EXEMPTIONS_FILENAME)
		self.similarity_exemptions=path.join(data_path,self.SIMILARITY_EXEMPTIONS_FILENAME)

		self.karaoke=get_paths(config, "karaoke", True)
		self.music=get_paths(config, "music", False)

		if path.exists(self.requests):
			remove(self.requests)


from karaokemanager.display_functions import pad_or_ellipsize, COLUMN_SEPARATOR
from colorama import Fore, Style

# Class describing a music file
class MusicFile:
	path = None
	title = None
	artist = None
	lower_title = None
	lower_artist = None
	search_string = None

	def __init__(self, path, artist, title):
		self.path = path
		self.title = title
		self.artist = artist
		self.lower_title = title.lower()
		self.lower_artist = artist.lower()
		self.search_string = f"{title} {artist}".lower()

	def matches(self, search_string):
		bits = search_string.strip().split()
		match_count = 0
		for bit in bits:
			if bit.lower() in self.search_string:
				match_count = match_count+1
		return match_count == len(bits)

	def get_text_with_specific_artist_and_title_lengths(self, max_artist_length, max_title_length):
		artist_text = pad_or_ellipsize(f"{self.artist} ", max_artist_length, ".")
		title_text = pad_or_ellipsize(self.title, max_title_length, " ")
		return f"{Fore.CYAN}{Style.BRIGHT}{artist_text}{Style.RESET_ALL}{COLUMN_SEPARATOR}{Fore.GREEN}{Style.BRIGHT}{title_text}{Style.RESET_ALL}"

	def get_text(self, max_length):
		max_length=max_length-len(COLUMN_SEPARATOR)
		max_artist_length=int(max_length/3)
		max_title_length=max_length-max_artist_length
		return self.get_text_with_specific_artist_and_title_lengths(max_artist_length,max_title_length)

	def get_option_text(self, max_length):
		return self.get_text(max_length)

	def get_song_list_text(self, max_artist_length, max_title_length):
		return self.get_text_with_specific_artist_and_title_lengths(max_artist_length, max_title_length)





from colorama import Fore, Style

# Helper class representing an on-screen column of singers, in the singers
# display.
class SingerColumn:
	singers = []
	index_start = 0
	column_width = 0

	def __init__(self, index_start, column_singers):
		self.index_start = index_start
		self.singers = column_singers
		max_singer_len = 8  # "nn: (n) "
		max_song_count = max(len(singer.songs) for singer in column_singers)
		max_name_length = max(len(singer.name) for singer in column_singers)
		max_singer_len += max_name_length
		if max_song_count > 9:
			max_singer_len += 1
		self.column_width = max_singer_len

	def get_row_text(self, row):
		if len(self.singers) > row:
			index = self.index_start+row
			index_text = f"{index}"
			if index < 10:
				index_text = f" {index_text}"
			singer = self.singers[row]
			song_count = len(singer.songs)
			plain_index_text = index_text
			index_text = f"{Fore.YELLOW}{Style.BRIGHT}{index_text}{Style.RESET_ALL}"
			name_text = ''
			if song_count > 0:
				name_text += f"{Fore.WHITE}{Style.BRIGHT}"
			else:
				name_text += f"{Fore.MAGENTA}{Style.NORMAL}"
			plain_name_text = singer.name
			name_text += singer.name
			row_text = f"{index_text}: {name_text}{Fore.CYAN}({song_count}){Style.RESET_ALL}"
			plain_row_text = f"{plain_index_text}: {plain_name_text}({song_count})"
			size_diff = self.column_width-len(plain_row_text)
			padding = " "*size_diff
			return row_text+padding
		return ""





import threading
import random
from time import sleep

class SuggestionGenerator:
	# Thread that will periodically choose a random karaoke file and write it to the suggestion file.
	suggestor_thread = None
	# Flag to stop the suggestion thread.
	stop_suggestions = False
	# Output path
	path=None

	def __init__(self,config):
		self.path=config.paths.random_suggestion

	# Function to stop the suggestion thread.
	def stop_suggestion_thread(self):
		if not self.suggestor_thread is None:
			self.stop_suggestions = True
			self.suggestor_thread.join()
		self.stop_suggestions = False

	# Function to start the suggestion thread.
	def start_suggestion_thread(self, karaoke_dictionary):
		self.stop_suggestion_thread()
		self.suggestor_thread = threading.Thread(
			target=self.random_song_suggestion_generator_thread, args=[karaoke_dictionary,])
		self.suggestor_thread.daemon = True
		self.suggestor_thread.start()

	# Thread that periodically writes a random karaoke suggestion to a file.
	def random_song_suggestion_generator_thread(self,dictionary):
		random.seed()
		counter = 0
		while not self.stop_suggestions:
			if counter == 0:
				counter = 20
				if any(dictionary):
					artistKeys = list(dictionary.keys())
					randomArtistIndex = random.randrange(len(dictionary))
					artistString = artistKeys[randomArtistIndex]
					artistDict = dictionary[artistString]
					randomSongIndex = random.randrange(len(artistDict))
					songKeys = list(artistDict.keys())
					songString = songKeys[randomSongIndex]
					suggestionString = f"{artistString}\n{songString}\n"
					try:
						with open(self.path, mode="w", encoding="utf-8") as f:
							f.writelines(suggestionString)
					except PermissionError:
						pass
			else:
				counter -= 1
			sleep(0.5)




from karaokemanager.music_file import MusicFile
from karaokemanager.display_functions import COLUMN_SEPARATOR, pad_or_ellipsize
from colorama import Fore, Style

# Class describing a karaoke file
class KaraokeFile(MusicFile):
	company = None

	def __init__(self, path, artist, title, company):
		MusicFile.__init__(self, path, artist, title)
		self.company = company

	def get_option_text(self, max_length):
		company_length=int(max_length/5)
		remaining_length=max_length-(company_length+len(COLUMN_SEPARATOR))
		company_text = pad_or_ellipsize(self.company, company_length, " ")
		song_text=MusicFile.get_option_text(self, remaining_length) # Company text, plus a space
		return f"{song_text}{COLUMN_SEPARATOR}{Style.BRIGHT}{Fore.YELLOW}{company_text}{Style.RESET_ALL}"



from karaokemanager.error import Error

# Class representing the list of songs that a singer has chosen.
class SongList:
	name = ''
	songs = []

	def __init__(self, name):
		self.name = name
		self.songs = []

	def write_to_state_file(self, file):
		file.writelines(self.name+"\n")
		for song in self.songs:
			song.write_to_state_file(file)

	def write_to_queue_file(self, file):
		if len(self.songs) == 0:
			indent = "\t"
		else:
			indent = ""
		file.writelines(indent+self.name+"\n")

	def get_song_index_from_id(self, id, end_allowed, errors):
		if id.isdigit():
			song_index = int(id)
			if song_index <= 0 or song_index > len(self.songs):
				errors.append(Error(f"Song index out of bounds: it must be between 1 and {len(self.songs)}"))
				return None
			return song_index-1
		id = id.lower()
		if len(self.songs) > 0:
			if id == "next" or id == "n":
				return 0
			if end_allowed:
				if id == "end" or id == "e":
					return len(self.songs)+1
		matches = []
		for i, song in enumerate(self.songs):
			if id in song.file.title.lower() or id in song.file.artist.lower():
				matches.append(i)
		if len(matches) == 0:
			errors.append(Error(f"No match found for given song ID \"{id}\""))
			return None
		if len(matches) > 1:
			errors.append(Error(f"Multiple matches found for given song ID \"{id}\""))
			return None
		return matches[0]

	def move_song(self, song_to_move_id, song_to_move_before_id, errors):
		matched_song_to_move_index = self.get_song_index_from_id(song_to_move_id, False, errors)
		if not matched_song_to_move_index is None:
			matched_song_to_move_before_index = self.get_song_index_from_id(song_to_move_before_id, True, errors)
			if not matched_song_to_move_before_index is None:
				song_to_move = self.songs[matched_song_to_move_index]
				del self.songs[matched_song_to_move_index]
				self.insert_song(matched_song_to_move_before_index, song_to_move)

	def insert_song(self, position, song):
		self.songs.insert(position, song)

	def get_maximum_lengths(self, limit):
		sliced_songs=self.songs[:limit]
		if any(sliced_songs):
			return max(map(lambda song: len(song.file.artist), sliced_songs)), max(map(lambda song: len(song.file.title), sliced_songs))
		return 0, 0


from colorama import Fore, Style
from os import get_terminal_size
import sys
from karaokemanager.commands import CommandType, parse_command
from karaokemanager.exemptions import Exemptions
from karaokemanager.state import State
from karaokemanager.library import Library, LibraryAnalysisType
from karaokemanager.karaoke_file import KaraokeFile
from karaokemanager.singer_column import SingerColumn
from karaokemanager.song_selector import select_song, show_song_list
from karaokemanager.display_functions import COLUMN_SEPARATOR, clear, pad_or_ellipsize
from karaokemanager.config import Config
from karaokemanager.error import Error, Info
from karaokemanager.suggestion_generator import SuggestionGenerator
from karaokemanager.song import Song

class KaraokeManager:
	# Default YAML config filename
	DEFAULT_CONFIG_FILENAME='.yaml'
	# Version number
	VERSION="1.0.4"
	# How many singers to show in each column?
	SINGERS_PER_COLUMN = 10
	# Minimum console width
	MIN_CONSOLE_WIDTH=120
	# Minimum console height
	MIN_CONSOLE_HEIGHT=30

	config=None
	console_size=None
	exemptions=None
	library=None

	def __init__(self):
		# Main execution loop
		console_size = get_terminal_size()
		if console_size.columns<KaraokeManager.MIN_CONSOLE_WIDTH:
			raise Exception(f"Karaoke Manager requires a console with a minimum of {KaraokeManager.MIN_CONSOLE_WIDTH} columns")
		if console_size.lines<KaraokeManager.MIN_CONSOLE_HEIGHT:
			raise Exception(f"Karaoke Manager requires a console with a minimum of {KaraokeManager.MIN_CONSOLE_HEIGHT} rows")
		self.console_size=console_size
		
		config_path=KaraokeManager.DEFAULT_CONFIG_FILENAME
		if len(sys.argv)>1:
			config_path=sys.argv[1]

		try:
			self.config=Config(config_path)
		except Exception as e:
			print(f"{Fore.RED}Error parsing the configuration file: {e}{Style.RESET_ALL}")
			exit(1)

		try:
			self.exemptions=Exemptions(self.config)
		except Exception as e:
			print(f"{Fore.RED}Error reading an exemptions file: {e}{Style.RESET_ALL}")
			exit(1)

		self.library = Library(self.exemptions)

	# Shows onscreen help when the user types "help"
	def show_help(self):
		clear()
		self.show_header()
		print(f"All command keywords can be shortened to the first letter.")
		print(f"IDs can be a number from the list, or a partial or complete name (case-insensitive).")
		print()
		print(f"{Fore.CYAN}{Style.BRIGHT}add,NewSingerName{Style.RESET_ALL} = add a new singer to the list")
		print(f"{Fore.CYAN}{Style.BRIGHT}insert,NewSingerName,next/SingerID{Style.RESET_ALL} = insert a new singer before an existing singer")
		print(f"{Fore.CYAN}{Style.BRIGHT}move,SingerID,next/end/SingerID{Style.RESET_ALL} = move a singer to another position in the list")
		print(f"{Fore.CYAN}{Style.BRIGHT}delete,SingerID{Style.RESET_ALL} = remove a singer from the list")
		print(f"{Fore.CYAN}{Style.BRIGHT}name,SingerID,NewName{Style.RESET_ALL} = rename a singer")
		print()
		print(f"{Fore.CYAN}{Style.BRIGHT}add,list/SingerID,SongTitle[,KeyChange]{Style.RESET_ALL} = add a song to a singer, or to current list")
		print(f"{Fore.CYAN}{Style.BRIGHT}insert,list/SingerID,next/SongID,SongTitle[,KeyChange]{Style.RESET_ALL} = insert a song into a list for a singer")
		print(f"{Fore.CYAN}{Style.BRIGHT}move,list/SingerID,SongID,next/end/SongID{Style.RESET_ALL} = move a song to another position in the list")
		print(f"{Fore.CYAN}{Style.BRIGHT}delete,list/SingerID,SongID{Style.RESET_ALL} = remove a song from a singer")
		print(f"{Fore.CYAN}{Style.BRIGHT}key,list/SingerID,SongID{Style.RESET_ALL} = set the key change on a song")
		print()
		print(f"{Fore.CYAN}{Style.BRIGHT}undo{Style.RESET_ALL} = undo previous command")
		print(f"{Fore.CYAN}{Style.BRIGHT}redo{Style.RESET_ALL} = undo previous undo")
		print()
		print(f"{Fore.CYAN}{Style.BRIGHT}list[,SingerID]{Style.RESET_ALL} = change active song list to show list of songs for specified singer (default is next singer)")
		print(f"{Fore.CYAN}{Style.BRIGHT}play[,SongID]{Style.RESET_ALL} = play song and cycle queue. Defaults to next song.")
		print(f"{Fore.CYAN}{Style.BRIGHT}filler[,SongID]{Style.RESET_ALL} = play song without cycling queue. Defaults to next song.")
		print(f"{Fore.CYAN}{Style.BRIGHT}?searchText{Style.RESET_ALL} = search karaoke tracks for given text")
		print(f"{Fore.CYAN}{Style.BRIGHT}??searchText{Style.RESET_ALL} = search music tracks for given text")
		print(f"{Fore.CYAN}{Style.BRIGHT}cue,SongTitle[,add]{Style.RESET_ALL} = cue up a background music track, optionally adding it to the playlist permanently.")
		print(f"{Fore.CYAN}{Style.BRIGHT}scan[,[quick]analyze]{Style.RESET_ALL} = rescan current folder for tracks, optionally analyzing for problems/duplicates.")
		print(f"{Fore.CYAN}{Style.BRIGHT}zap{Style.RESET_ALL} = clear everything and start from scratch")
		print(f"{Fore.CYAN}{Style.BRIGHT}help{Style.RESET_ALL} = show this screen")
		print(f"{Fore.CYAN}{Style.BRIGHT}quit{Style.RESET_ALL} = quit")
		try:
			input(f"{Fore.WHITE}{Style.BRIGHT}Press Enter to continue...{Style.RESET_ALL}")
		except EOFError:
			pass

	# Cues up a song
	def cue_song(self, params, feedback):
		if not any(params):
			feedback.append(Error("Not enough arguments. Expected song search string."))
		else:
			song = select_song(params[0], self.library.music_files, self.console_size)
			if not song is None:
				feedback.append(Info(f"Added \"{song.title}\" by {song.artist} to the requests queue."))
				try:
					with open(self.config.paths.requests_filename, mode="a", encoding="utf-8") as f:
						f.write(song.path+"\n")
				except PermissionError:
					feedback.append(Error("Failed to write to requests file."))
				if len(params)>1:
					if params[1]=="a" or params[1]=="add":
						try:
							with open(self.config.paths.background_music_list_path, mode="a", encoding="utf-8") as f:
								f.write(f"{song.artist} - {song.title}\n")
						except PermissionError:
							feedback.append(Error("Failed to append to background music playlist file."))

	# Prints the app header
	def show_header(self):
		print(f'{Fore.GREEN}{Style.BRIGHT}Karaoke Manager v{self.VERSION}{Style.RESET_ALL} ({len(self.library.karaoke_files)}/{len(self.library.music_files)} karaoke/music files found)')

	# Print the current list of singers. Groups them into columns.
	def show_singers(self, state):
		singers_with_requests = state.getSingersDisplayList()
		if not any(singers_with_requests):
			print("No singers.")
		else:
			columns_of_singers = []
			dictionary = [{'chunk': chunk, 'index': i} for i, chunk in enumerate(chunks(singers_with_requests, KaraokeManager.SINGERS_PER_COLUMN))]
			columns_of_singers = list(map(lambda item: SingerColumn((item['index']*KaraokeManager.SINGERS_PER_COLUMN)+1, item['chunk']), dictionary))
			for row in range(0, min(len(singers_with_requests), KaraokeManager.SINGERS_PER_COLUMN)):
				print(*map(lambda singerColumn: singerColumn.get_row_text(row), columns_of_singers))

	# Print the list of songs for the current singer, or whatever singer
	# has been flagged as the current "active list" singer.
	def show_songs(self, state, feedback):
		SONG_LIST_INDEX_LENGTH=4 # Two digits, colon and space
		active_singer = state.get_active_song_list_singer()
		if active_singer is None:
			print(f"{Fore.MAGENTA}No current singer selected.{Style.RESET_ALL}")
		else:
			is_song_list_singer_next = (active_singer == state.next_singer())
			if is_song_list_singer_next:
				name_color = f"{Fore.WHITE}"
			else:
				name_color = f"{Fore.MAGENTA}"
			print(f"{Fore.WHITE}Showing song list for {name_color}{Style.BRIGHT}{active_singer.name}{Style.RESET_ALL}:")
			rows_used=min(len(state.singers),KaraokeManager.SINGERS_PER_COLUMN)
			rows_used+=3 # Blank rows (one above and below singers list, and one after this song list)
			rows_used+=2 # App header, and caption for this song list
			rows_used+=2 # Upcoming prompt, and line for user to type on
			rows_used+=len(feedback) # Any errors/messages we need to show the user
			rows_available=self.console_size.lines-rows_used
			songs_in_list=len(active_singer.songs)
			overflow=songs_in_list-rows_available
			if overflow>0:
				rows_available-=1 # Have to show one fewer than that, so that we can show "+n more"

			max_artist_length, max_title_length=active_singer.get_maximum_lengths(rows_available)
			# We need to leave room for the song index, plus any key change indicator (and a column separator before that)
			# The rest of the screen is available for song artist + title
			max_available_space=self.console_size.columns-(SONG_LIST_INDEX_LENGTH+len(COLUMN_SEPARATOR)+Song.KEY_CHANGE_INDICATOR_LENGTH)
			combined_length=max_artist_length+max_title_length+len(COLUMN_SEPARATOR)
			# If the longest artist + longest title bigger than the available space? Scale 'em down.
			if max_available_space<combined_length:
				reduction_factor=max_available_space/combined_length
				max_artist_length=int(max_artist_length*reduction_factor)
				max_title_length=int(max_title_length*reduction_factor)
			for i, song in enumerate(active_singer.songs[:rows_available]):
				song_index = f"{Fore.YELLOW}{Style.BRIGHT}{i+1}{Style.RESET_ALL}"
				if i < 9:
					song_index = f" {song_index}"
				print(f"{song_index}: {song.get_song_list_text(max_artist_length, max_title_length)}")
			if overflow>0:
				print(f"{Fore.GREEN}{Style.BRIGHT}... and {overflow} more.{Style.RESET_ALL}")

	def show_library_report(self):
		if any(self.library.unparseable_filenames) or any(self.library.missing_bgm_playlist_entries) or any(self.library.karaoke_analysis_results) or any(self.library.music_analysis_results) or any(self.library.karaoke_duplicates) or any (self.library.music_duplicates):
			scanCompleteMessage=pad_or_ellipsize("Scan complete.", self.console_size.columns)
			print(f"{Fore.WHITE}{Style.BRIGHT}{scanCompleteMessage}")
			print(f"{Fore.RED}{Style.BRIGHT}Bad filenames:{Style.RESET_ALL} {len(self.library.unparseable_filenames)}")
			print(f"{Fore.GREEN}{Style.BRIGHT}Ignored files:{Style.RESET_ALL} {len(self.library.ignored_files)}")
			print(f"{Fore.YELLOW}{Style.BRIGHT}Artist/title problems:{Style.RESET_ALL} {len(self.library.karaoke_analysis_results)+len(self.library.music_analysis_results)}")
			print(f"{Fore.CYAN}{Style.BRIGHT}Duplicate files:{Style.RESET_ALL} {len(self.library.karaoke_duplicates)+len(self.library.music_duplicates)}")
			print(f"{Fore.MAGENTA}{Style.BRIGHT}Missing playlist entries:{Style.RESET_ALL} {len(self.library.missing_bgm_playlist_entries)}")
			try:
				input("Press Enter to continue ...")
			except EOFError:
				pass

	def rebuild_library(self, analysis_type, feedback):
		self.library.build_song_lists(self.config, analysis_type, self.console_size, feedback)
		self.show_library_report()

	# Processes the given command.
	# Returns true if the command is to quit the app.
	def process_command(self, command, state, feedback):
		if command.command_type == CommandType.HELP:
			self.show_help()
		elif command.command_type == CommandType.QUIT:
			return None
		elif command.command_type == CommandType.ADD:
			state = state.add(command.params, self.library.karaoke_files, self.console_size, feedback)
		elif command.command_type == CommandType.INSERT:
			state = state.insert(command.params, self.library.karaoke_files, self.console_size, feedback)
		elif command.command_type == CommandType.MOVE:
			state = state.move(command.params, feedback)
		elif command.command_type == CommandType.DELETE:
			state = state.delete(command.params, feedback)
		elif command.command_type == CommandType.LIST:
			state = state.list(command.params, feedback)
		elif command.command_type == CommandType.UNDO:
			state = state.undo(feedback)
		elif command.command_type == CommandType.REDO:
			state = state.redo(feedback)
		elif command.command_type == CommandType.SCAN:
			self.rebuild_library(parse_library_analysis_type(command.params), feedback)
		elif command.command_type == CommandType.ZAP:
			state = state.clear()
		elif command.command_type == CommandType.NAME:
			state = state.rename_singer(command.params, feedback)
		elif command.command_type == CommandType.PLAY:
			state = state.play(command.params, True, self.config.driver, feedback)
		elif command.command_type == CommandType.FILLER:
			state = state.play(command.params, False, self.config.driver, feedback)
		elif command.command_type == CommandType.KEY:
			state = state.change_song_key(command.params, feedback)
		elif command.command_type == CommandType.CUE:
			self.cue_song(command.params, feedback)
		elif command.command_type == CommandType.SEARCH:
			show_song_list(command.params[0], self.library.karaoke_files, False, self.console_size)
		elif command.command_type == CommandType.MUSIC_SEARCH:
			show_song_list(command.params[0], self.library.music_files, False, self.console_size)
		return state

	# Asks the user for a command, and parses it
	def get_command(self, feedback):
		try:
			command = input(":")
		except EOFError:
			pass
		command = command.strip()
		if len(command) > 0:
			parsed_command, command_string = parse_command(command)
			if parsed_command is None:
				feedback.append(Error(f"Unknown command: \"{command_string}\""))
			return parsed_command
		return None

	# Shows any info/errors from the previous command
	def show_feedback(self, feedback):
		for message in feedback:
			message.print()

	def run(self):
		clear()
		feedback=[]
		suggestion_generator=SuggestionGenerator(self.config)
		self.rebuild_library(LibraryAnalysisType.NONE, feedback)
		suggestion_generator.start_suggestion_thread(self.library.karaoke_dictionary)
		state = State(self.config, self.library, feedback)
		while True:
			clear()
			state.save(feedback)
			self.show_header()
			print()
			self.show_singers(state)
			print()
			self.show_songs(state, feedback)
			print()
			self.show_feedback(feedback)
			feedback=[]
			print(f"Enter command, or type {Style.BRIGHT}help{Style.NORMAL} to see list of commands.")
			command = self.get_command(feedback)
			if not command is None:
				state=self.process_command(command, state, feedback)
				if command.command_type==CommandType.SCAN:
					suggestion_generator.start_suggestion_thread(self.library.karaoke_dictionary)
				if state is None:
					break
				
		clear()
		suggestion_generator.stop_suggestion_thread()

# Splits a collection into smaller collections of the given size
def chunks(l, n):
	for i in range(0, len(l), n):
		yield l[i:i + n]

def parse_library_analysis_type(params):
	if any(params):
		if params[0] == "quickanalyze" or params[0] == "q":
			return LibraryAnalysisType.QUICK
		if params[0] == "analyze" or params[0] == "a":
			return LibraryAnalysisType.FULL
	return LibraryAnalysisType.NONE


from os import path
from copy import deepcopy
from karaokemanager.singer import Singer
from karaokemanager.song_selector import select_song
from karaokemanager.song import Song
from karaokemanager.error import Error

# Current state of the app.
class State:
	singers = None
	active_song_list_singer_name = None
	next_state = None
	prev_state = None
	state_path = None
	queue_path = None	

	def __init__(self, config, library, errors):
		self.state_path = config.paths.state
		self.queue_path = config.paths.singers_queue
		self.load(library.karaoke_files, errors)

	# Load the last saved state.
	def load(self, karaokeFiles, errors):
		self.singers = []
		if path.isfile(self.state_path):
			with open(self.state_path, mode="r", encoding="utf-8") as f:
				lines = f.readlines()
				current_singer = None
				for line in lines:
					if line.startswith("\t"):
						if not current_singer is None:
							lineBits = line.strip().split("|")
							karaoke_file = get_karaoke_file_from_path(lineBits[0], karaokeFiles)
							key_change = get_key_change_value(lineBits[1].strip(), errors)
							if key_change!=-99:
								if not karaoke_file is None:
									current_singer.songs.append(Song(karaoke_file, key_change))
					else:
						if not current_singer is None:
							self.singers.append(current_singer)
						if len(line) > 0:
							current_singer = Singer(line.strip())
				if not current_singer is None:
					self.singers.append(current_singer)

	def save(self, errors):
		try:
			with open(self.state_path, mode="w", encoding="utf-8") as f:
				for singer in self.singers:
					singer.write_to_state_file(f)
		except PermissionError:
			errors.append(Error("Failed to write state file."))
		try:
			with open(self.queue_path, mode="w", encoding="utf-8") as f:
				for singer in self.singers:
					singer.write_to_queue_file(f)
		except PermissionError:
			errors.append(Error("Failed to write singer names file."))

	def getSingersDisplayList(self):
		singersCopy = self.singers.copy()
		return singersCopy

	# Make a deep copy of the current state, ready for alteration.
	# We create a chain of previous/next states for undo/redo.
	def mutate(self):
		new_state = deepcopy(self)
		self.next_state = new_state
		new_state.prev_state = self
		return new_state

	def get_singer_index_from_id(self, id, end_allowed, req_allowed, errors):
		if id.isdigit():
			singer_index = int(id)
			if singer_index <= 0 or singer_index > len(self.singers):
				errors.append(Error(f"Singer index out of bounds: it must be between 1 and {len(self.singers)}"))
				return None
			return singer_index-1
		id = id.lower()
		if len(self.singers) > 0:
			if id == "next" or id == "n":
				return 0
			if id == "end" or id == "e":
				return len(self.singers)
		if id == "list" or id == "l":
			if not self.active_song_list_singer_name is None:
				id = self.active_song_list_singer_name.lower()
			else:
				next_singer = self.next_singer()
				if not next_singer is None:
					id = next_singer.name.lower()
				else:
					errors.append(Error("There is no active song list to add to."))
					return None
		for i, singer in enumerate(self.singers):
			if singer.name.lower() == id:
				return i
		matches = []
		for i, singer in enumerate(self.singers):
			if id in singer.name.lower():
				matches.append(i)
		if len(matches) == 0:
			errors.append(Error(f"No match found for given singer ID \"{id}\""))
			return None
		if len(matches) > 1:
			errors.append(Error(f"Multiple matches found for given singer ID \"{id}\""))
			return None
		return matches[0]

	def get_singer_from_id(self, id, end_allowed, req_allowed, errors):
		singer_index = self.get_singer_index_from_id(id, end_allowed, req_allowed, errors)
		if singer_index is None:
			return None
		return self.singers[singer_index]

	def add_new_singer(self, singer_name, index, errors):
		same_name_singers = [
			singer for singer in self.singers if singer.name.lower() == singer_name.lower()]
		if len(same_name_singers) > 0:
			errors.append(Error(f"Singer with name \"{singer_name}\" already exists."))
		else:
			new_state = self.mutate()			
			new_state.singers.insert(index, Singer(singer_name))
			return new_state
		return self

	def delete_singer(self, singer_id, errors):
		matched_singer_index = self.get_singer_index_from_id(singer_id, False, False, errors)
		if not matched_singer_index is None:
			new_state = self.mutate()
			del new_state.singers[matched_singer_index]
			return new_state
		return self

	def delete_song_from_singer(self, singer_id, song_id, errors):
		matched_singer_index = self.get_singer_index_from_id(singer_id, False, False, errors)
		if not matched_singer_index is None:
			matched_singer = self.singers[matched_singer_index]
			matched_song_index = matched_singer.get_song_index_from_id(song_id, False, errors)
			if not matched_song_index is None:
				new_state = self.mutate()
				del new_state.singers[matched_singer_index].songs[matched_song_index]
				return new_state
		return self

	def insert_new_singer(self, singer_name, singer_id, errors):
		matched_singer_index = self.get_singer_index_from_id(singer_id, True, False, errors)
		if not matched_singer_index is None:
			return self.add_new_singer(singer_name, matched_singer_index, errors)
		return self

	def move_singer(self, singer_to_move_id, singer_id, errors):
		matched_singer_to_move_index = self.get_singer_index_from_id(singer_to_move_id, False, False, errors)
		if not matched_singer_to_move_index is None:
			matched_singer_index = self.get_singer_index_from_id(singer_id, True, False, errors)
			if not matched_singer_index is None:
				new_state = self.mutate()
				singer_to_move = new_state.singers[matched_singer_to_move_index]
				del new_state.singers[matched_singer_to_move_index]
				new_state.singers.insert(matched_singer_index, singer_to_move)
				return new_state
		return self

	def add_song_for_singer(self, singer_id, song_title, key_change, index, karaoke_files, console_size, errors):
		key_change_value = get_key_change_value(key_change, errors)
		if key_change_value != -99:
			matched_singer = self.get_singer_from_id(singer_id, False, True, errors)
			if not matched_singer is None:
				karaoke_song = select_song(song_title, karaoke_files, console_size)
				if not karaoke_song is None:
					if index == -1:
						index = len(matched_singer.songs)
					new_state = self.mutate()
					matched_singer = new_state.get_singer_from_id(singer_id, False, True, errors)
					matched_singer.insert_song(index, Song(karaoke_song, key_change_value))
					return new_state
		return self

	def insert_song_for_singer(self, singer_id, song_id, song_title, key_change, karaoke_files, console_size, errors):
		matched_singer = self.get_singer_from_id(singer_id, False, True, errors)
		if not matched_singer is None:
			matched_song_index = matched_singer.get_song_index_from_id(song_id, True, errors)
			if not matched_song_index is None:
				return self.add_song_for_singer(singer_id, song_title, key_change, matched_song_index, karaoke_files, console_size, errors)
		return self

	def move_song(self, singer_id, song_to_move_id, song_to_move_before_id, errors):
		matched_singer = self.get_singer_from_id(singer_id, False, True, errors)
		if not matched_singer is None:
			matched_song_to_move_index = matched_singer.get_song_index_from_id(song_to_move_id, False, errors)
			if not matched_song_to_move_index is None:
				matched_song_to_move_before_index = matched_singer.get_song_index_from_id(song_to_move_before_id, True, errors)
				if not matched_song_to_move_before_index is None:
					new_state = self.mutate()
					matched_singer = new_state.get_singer_from_id(singer_id, False, True, errors)
					matched_singer.move_song(song_to_move_id, song_to_move_before_id, errors)
					return new_state
		return self

	def add(self, params, karaoke_files, console_size, errors):
		param_count = len(params)
		if param_count == 0:
			errors.append(Error("Not enough arguments. Expected name of new singer, or existing singer name/index."))
		elif param_count == 1:
			return self.add_new_singer(params[0], len(self.singers), errors)
		else:
			key_change = None
			if param_count > 2:
				key_change = params[2]
			return self.add_song_for_singer(params[0], params[1], key_change, -1, karaoke_files, console_size, errors)
		return self

	def insert(self, params, karaoke_files, console_size, errors):
		param_count = len(params)
		if param_count < 2:
			errors.append(Error("Not enough arguments. Expected name of new singer, or existing singer name/index."))
		elif param_count == 2:
			return self.insert_new_singer(params[0], params[1], errors)
		elif param_count > 2:
			keyChange = None
			if param_count > 3:
				keyChange = params[3]
			return self.insert_song_for_singer(params[0], params[1], params[2], keyChange, karaoke_files, console_size, errors)
		return self

	def move(self, params, errors):
		param_count = len(params)
		if param_count < 2:
			errors.append(Error("Not enough arguments. Expected ID of singer."))
		elif param_count == 2:
			# Move a singer in the list
			return self.move_singer(params[0], params[1], errors)
		elif param_count > 2:
			# Move a song in a singer's list
			return self.move_song(params[0], params[1], params[2], errors)
		return self

	def delete(self, params, errors):
		param_count = len(params)
		if param_count < 1:
			errors.append(Error("Not enough arguments. Expected ID of singer."))
		elif param_count == 1:
			# Delete a singer
			return self.delete_singer(params[0], errors)
		elif param_count > 1:
			# Delete a song from a singer
			return self.delete_song_from_singer(params[0], params[1], errors)
		return self

	def undo(self, errors):
		if self.prev_state is None:
			errors.append(Error("No undo history available."))
		else:
			return self.prev_state

	def redo(self, errors):
		if self.next_state is None:
			errors.append(Error("No redo history available."))
		else:
			return self.next_state

	def rename_singer(self, params, errors):
		if len(params) < 2:
			errors.append(Error("Not enough arguments. Expected singer ID and new name."))
		else:
			singer = self.get_singer_from_id(params[0], False, False, errors)
			new_name = params[1]
			if not singer is None:
				for singer in self.singers:
					if singer.name.lower() == new_name:
						errors.append(Error(f"Name \"{new_name}\" already exists."))
						return
				new_state = self.mutate()
				singer = new_state.get_singer_from_id(params[0], False, False, errors)
				singer.name = new_name
				return new_state
		return self

	def clear(self):
		new_state = self.mutate()
		new_state.singers = []
		new_state.active_song_list_singer_name = None
		return new_state

	def next_singer(self):
		for singer in self.singers:
			if len(singer.songs) > 0:
				return singer
		return None

	def set_song_list_to_next_singer(self):
		self.active_song_list_singer_name = None

	def set_song_list_to_singer(self, singer):
		self.active_song_list_singer_name = singer.name

	def list(self, params, errors):
		if len(params) == 0:
			next_state = self.mutate()
			next_state.set_song_list_to_next_singer()
			return next_state
		else:
			singer = self.get_singer_from_id(params[0], False, True, errors)
			if not singer is None:
				next_state = self.mutate()
				next_state.set_song_list_to_singer(singer)
				return next_state

	def get_active_song_list_singer(self):
		if self.active_song_list_singer_name is None:
			return self.next_singer()
		active_singer = [
			singer for singer in self.singers if singer.name == self.active_song_list_singer_name]
		if len(active_singer) == 0:
			self.active_song_list_singer_name = None
			return self.next_singer()
		return active_singer[0]

	def change_song_key(self, params, errors):
		if len(params) < 3:
			errors.append(Error("Not enough arguments. Expected singer ID, song ID, and new key change value."))
		else:
			singer = self.get_singer_from_id(params[0], False, False, errors)
			if not singer is None:
				song_index = singer.get_song_index_from_id(params[1], False, errors)
				if not song_index is None:
					key_change_string = params[2]
					if key_change_string == "0":
						key_change_string = None
					key_change_value = get_key_change_value(key_change_string, errors)
					if key_change_value != -99:
						new_state = self.mutate()
						singer = new_state.get_singer_from_id(params[0], False, False, errors)
						if not singer is None:
							song_index = singer.get_song_index_from_id(params[1], False, errors)
							if not song_index is None:
								singer.songs[song_index].key_change = key_change_value
								return new_state
		return self

	def play(self, params, cycle_queue, driver, errors):
		if len(params) > 0:
			song = params[0]
		else:
			song = "next"
		if not self.next_singer() is None:
			new_state = self.mutate()
			next_singer = new_state.next_singer()
			song_to_play_index = next_singer.get_song_index_from_id(song, False, errors)
			if not song_to_play_index is None:
				song = next_singer.songs[song_to_play_index]
				file_to_start = next_singer.songs[song_to_play_index].file.path
				if cycle_queue:
					new_state.singers.remove(next_singer)
					new_state.singers.append(next_singer)
				del next_singer.songs[song_to_play_index]
				driver.play_karaoke_file(file_to_start, song.key_change, errors)
				return new_state
		else:
			errors.append(Error("There are no singers with songs available."))
		return self

# Given a path, get the karaoke file that matches it.
# Used when restoring state on startup to check that a cued-up song still
# exists.				
def get_karaoke_file_from_path(path, karaokeFiles):
	path = path.lower()
	matches = [
		karaokeFile for karaokeFile in karaokeFiles if karaokeFile.path.lower() == path]
	if len(matches) == 1:
		return matches[0]
	return None

# Parses a keychange string.
def get_key_change_value(key_change, errors):
	if not key_change is None and not key_change=="0":
		if len(key_change) == 2:
			if key_change[0] == '+' or key_change[0] == '-':
				value = key_change[1:2]
				if value.isdigit():
					int_value = int(value)
					if int_value <= Song.MAX_KEY_CHANGE and int_value > 0:
						if key_change[0] == '-':
							return -int_value
						return int_value
		errors.append(Error("Invalid key change, should in format \"+N\" or \"-N\", where N is a value between 1 and 5."))
		return -99
	return 0


from enum import Enum, auto

# List of commands as enums
class CommandType(Enum):
	ADD = auto()
	INSERT = auto()
	MOVE = auto()
	DELETE = auto()
	LIST = auto()
	PLAY = auto()
	SCAN = auto()
	ZAP = auto()
	HELP = auto()
	QUIT = auto()
	UNDO = auto()
	REDO = auto()
	SEARCH = auto()
	NAME = auto()
	KEY = auto()
	FILLER = auto()
	MUSIC_SEARCH = auto()
	CUE = auto()

# Command definition class. A combination of the command enum
# and the string that is associated with the command.
class CommandDefinition:
	command_type = None
	command_string = None

	def __init__(self, command_type, command_string):
		self.command_type = command_type
		self.command_string = command_string

# List of command definitions
commands = [
	CommandDefinition(CommandType.ADD, "add"),
	CommandDefinition(CommandType.INSERT, "insert"),
	CommandDefinition(CommandType.MOVE, "move"),
	CommandDefinition(CommandType.DELETE, "delete"),
	CommandDefinition(CommandType.LIST, "list"),
	CommandDefinition(CommandType.PLAY, "play"),
	CommandDefinition(CommandType.SCAN, "scan"),
	CommandDefinition(CommandType.ZAP, "zap"),
	CommandDefinition(CommandType.HELP, "help"),
	CommandDefinition(CommandType.UNDO, "undo"),
	CommandDefinition(CommandType.REDO, "redo"),
	CommandDefinition(CommandType.QUIT, "quit"),
	CommandDefinition(CommandType.KEY, "key"),
	CommandDefinition(CommandType.NAME, "name"),
	CommandDefinition(CommandType.FILLER, "filler"),
	CommandDefinition(CommandType.CUE, "cue"),
	CommandDefinition(CommandType.SEARCH, "?"),
	CommandDefinition(CommandType.MUSIC_SEARCH, "??")
]

MUSIC_SEARCH_STRING="??"
KARAOKE_SEARCH_STRING="?"

# Result of a parsed command
class Command:
	command_type = None
	# The series of additional parameters that were entered after the command string
	# e.g. "add,Steve,Don't Look Back In Anger" would have two params: "Steve", and
	# "Don't Look Back In Anger"
	params = []

	def __init__(self, command_type, params):
		self.command_type = command_type
		self.params = params

# Attempts to parse the command type by checking the command string that was entered.
def parse_command_type(command_string):
	command_string = command_string.lower()
	for command in commands:
		if command_string == command.command_string or command_string == command.command_string[0]:
			return command.command_type
	return None

# Parses the given command line into a Command object
def parse_command(command_string):
	# Special case = search
	if command_string[0:2] == MUSIC_SEARCH_STRING:
		return Command(CommandType.MUSIC_SEARCH, [command_string[2:]]), MUSIC_SEARCH_STRING
	if command_string[0] == KARAOKE_SEARCH_STRING:
		return Command(CommandType.SEARCH, [command_string[1:]]), KARAOKE_SEARCH_STRING
	command_bits = command_string.split(',')
	for i, command_bit in enumerate(command_bits):
		command_bits[i] = command_bit.strip()
	command_type = parse_command_type(command_bits[0])
	if command_type is None:
		return None, command_bits[0]
	else:
		return Command(command_type, command_bits[1:]), command_bits[0]



from karaokemanager.song_list import SongList

# Class describing a singer, and their current list of selected songs.
class Singer(SongList):
	def __init__(self, name):
		SongList.__init__(self, name)

	def write_to_state_file(self, file):
		SongList.write_to_state_file(self, file)

	def write_to_queue_file(self, file):
		SongList.write_to_queue_file(self, file)

	def get_song_index_from_id(self, id, end_allowed, errors):
		return SongList.get_song_index_from_id(self, id, end_allowed, errors)

	def move_song(self, song_to_move_id, song_to_move_before_id, errors):
		return SongList.move_song(self, song_to_move_id, song_to_move_before_id, errors)

	def insert_song(self, position, song):
		SongList.insert_song(self, position, song)




from karaokemanager.display_functions import clear
from colorama import Fore, Style
from karaokemanager.music_file import MusicFile

# Song selector function. Shows the list of "songs" to the user and asks them to pick one,
# cancel, or enter more text to narrow the search.
def select_song(searchString, songs, console_size):
	return show_song_list(searchString, songs, True, console_size)

# Asks the user to choose a song from the displayed list, and awaits a valid
# response.
def get_song_choice(message, options, blank_allowed, selection_allowed):
	invalid_choice = True
	while invalid_choice:
		invalid_choice = False
		try:
			song_num = input(message)
		except EOFError:
			pass
		song_num = song_num.strip()
		if song_num.lower().startswith("x"):
			return None
		if song_num.isdigit() and selection_allowed:
			parsed_int = int(song_num)-1
			if parsed_int < 0 or parsed_int > len(options):
				invalid_choice = True
				print(
					f"{Fore.RED}{Style.BRIGHT}Invalid choice, try again.{Style.RESET_ALL}")
			else:
				return options[parsed_int]
		else:
			if len(song_num) == 0:
				if blank_allowed:
					return ""
				else:
					invalid_choice = True
					print(
						f"{Fore.RED}{Style.BRIGHT}Invalid choice, try again.{Style.RESET_ALL}")
			else:
				return song_num

# Displays the list of songs and optionally prompts for a response.
def show_song_list(search_string, files, selection_allowed, console_size):
	search_again = True
	if selection_allowed:
		optional_selection_text = ", song # to select,"
	else:
		optional_selection_text = ""
	while search_again:
		search_again = False
		options = [
			song_file for song_file in files if song_file.matches(search_string)]
		if len(options) == 1 and selection_allowed:
			return options[0]
		if len(options) == 0:
			print(f"{Fore.RED}{Style.BRIGHT}No results found for \"{search_string}\" ...{Style.RESET_ALL}")
			try:
				input("Press Enter to continue.")
			except EOFError:
				pass
			return None
		clear()
		shown_count = 0
		total_shown_count = 0
		start_count = 1
		for i, option in enumerate(options):
			index_string = f"{i+1}"
			padding = (3-len(index_string))*" "
			index_string = padding+index_string
			print(f"{Fore.YELLOW}{Style.BRIGHT}{index_string}{Style.RESET_ALL}: {option.get_option_text(console_size.columns-5)}")
			shown_count += 1
			total_shown_count += 1
			if shown_count == console_size.lines-2 or i == len(options)-1:
				print(f"Showing results {Fore.WHITE}{Style.BRIGHT}{start_count}-{start_count+(shown_count-1)}{Style.RESET_ALL} of {Fore.WHITE}{Style.BRIGHT}{len(options)}{Style.RESET_ALL} for {Fore.YELLOW}{Style.BRIGHT}\"{search_string}\"{Style.RESET_ALL}.")
				if len(options) > total_shown_count:
					song_number = get_song_choice(f"Press Enter for more, or type x to cancel{optional_selection_text} or additional search text: ", options, True, selection_allowed)
					if song_number is None:
						return None
					elif song_number == "":
						start_count += shown_count
						shown_count = 0
						continue
					elif isinstance(song_number, MusicFile):
						return song_number
					else:
						search_string = search_string+" "+song_number
						search_again = True
						break
		if not search_again:
			if not selection_allowed:
				try:
					input("Press Enter to continue ...")
				except EOFError:
					pass
				return None
			song_number = get_song_choice("Enter a song # to select, x to cancel, or additional search text: ", options, False, selection_allowed)
			if song_number is None:
				return None
			elif isinstance(song_number, MusicFile):
				return song_number
			else:
				search_string = search_string+" "+song_number
				search_again = True


from colorama import Fore, Style
from karaokemanager.display_functions import COLUMN_SEPARATOR

# Class representing a chosen song, with any requested key change
class Song:
	file = None
	key_change = None

	MAX_KEY_CHANGE=5 # To prevent things getting stupid

	KEY_CHANGE_INDICATOR_LENGTH=4 # Two parentheses, a sign, a digit

	def __init__(self, file, key_change):
		self.file = file
		self.key_change = key_change

	def write_to_state_file(self, file):
		sign=""
		if(self.key_change>0):
			sign="+"
		file.write(f"\t{self.file.path}|{sign}{self.key_change}\n")

	def get_song_list_text(self, max_artist_length, max_title_length):
		key_change_text=""
		if self.key_change!=0:
			sign=""
			if self.key_change>0:
				sign="+"
			key_change_text = f"{COLUMN_SEPARATOR}{Fore.MAGENTA}{Style.BRIGHT}({sign}{self.key_change}){Style.RESET_ALL}"
		song_text=self.file.get_song_list_text(max_artist_length, max_title_length)
		return f"{song_text}{key_change_text}"



from os import name, system

COLUMN_SEPARATOR="  "

# Pads the given string with spaces to make it reach the desired length.
# If it is already too long, it will be truncated and ellipsised.
def pad_or_ellipsize(str, length, padStr=" "):
	if len(str) > length:
		str = str.strip()
		if len(str) > length:
			str = f"{str[0:length-3].strip()}..."
	return str+((length-len(str))*padStr)

# Clear the screen. Have to use system commands for this.
def clear():
	# for windows
	if name == 'nt':
		_ = system('cls')
	# for mac and linux(here, os.name is 'posix')
	else:
		_ = system('clear')


from setuptools import setup, find_packages
from karaokemanager.karaoke_manager import KaraokeManager

with open('README.md') as readme_file:
    README = readme_file.read()

with open('HISTORY.md') as history_file:
    HISTORY = history_file.read()

setup_args = dict(
	name='karaokemanager',
	version=KaraokeManager.VERSION,
	description='Karaoke session management utility',
	long_description_content_type="text/markdown",
	long_description=README + '\n\n' + HISTORY,
	license='MIT',
	packages=find_packages(),
	entry_points={
		"console_scripts": [
				"karaokemanager=karaokemanager.__main__:main",
		]
	},
	author='Steven Frew',
	author_email='steven.fullhouse@gmail.com',
	keywords=['karaokemanager', 'karaoke'],
	url='https://github.com/peeveen/karaokemanager',
	download_url='https://pypi.org/project/karaokemanager/'
)

install_requires = ['colorama', 'pyyaml', 'textdistance']

if __name__ == '__main__':
    setup(**setup_args, install_requires=install_requires)


import re
from collections import defaultdict

class FilePattern:
	regular_expression=None
	extensions=[]
	groups=[]
	def __init__(self, extensions, regular_expression, groups):
		self.extensions=list(map(lambda ext: ext.lower(),extensions))
		self.regular_expression=re.compile(regular_expression)
		self.groups=groups

	def extension_matches(self, ext):
		return ext.lower() in self.extensions
	
	def parse_filename(self, filename):
		group_map=defaultdict()
		match=self.regular_expression.match(filename)
		if not match is None:
			for i,group_name in enumerate(self.groups):
				if (not group_name is None) and group_name!="":
					matched_string=match.group(i+1)
					if (not matched_string is None) and matched_string!="":
						group_map[group_name]=matched_string
		return group_map
